"""Common functions between v1 transfer and liquid-class-based transfer."""
import enum
from typing import Iterable, Generator, Tuple, TypeVar, Literal


class TransferTipPolicyV2(enum.Enum):
    ONCE = "once"
    NEVER = "never"
    ALWAYS = "always"
    PER_SOURCE = "per source"


TransferTipPolicyV2Type = Literal["once", "always", "per source", "never"]

Target = TypeVar("Target")


def check_valid_volume_parameters(
    disposal_volume: float, air_gap: float, max_volume: float
) -> None:
    if air_gap >= max_volume:
        raise ValueError(
            "The air gap must be less than the maximum volume of the pipette"
        )
    elif disposal_volume >= max_volume:
        raise ValueError(
            "The disposal volume must be less than the maximum volume of the pipette"
        )
    elif disposal_volume + air_gap >= max_volume:
        raise ValueError(
            "The sum of the air gap and disposal volume must be less than"
            " the maximum volume of the pipette"
        )


def expand_for_volume_constraints(
    volumes: Iterable[float],
    targets: Iterable[Target],
    max_volume: float,
) -> Generator[Tuple[float, "Target"], None, None]:
    """Split a sequence of proposed transfers if necessary to keep each
    transfer under the given max volume.
    """
    # A final defense against an infinite loop.
    # Raising a proper exception with a helpful message is left to calling code,
    # because it has more context about what the user is trying to do.
    assert max_volume > 0
    for volume, target in zip(volumes, targets):
        while volume > max_volume * 2:
            yield max_volume, target
            volume -= max_volume

        if volume > max_volume:
            volume /= 2
            yield volume, target
        yield volume, target



"""Aspirate in place command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union
from typing_extensions import Literal

from opentrons.hardware_control import HardwareControlAPI

from .pipetting_common import (
    PipetteIdMixin,
    AspirateVolumeMixin,
    FlowRateMixin,
    BaseLiquidHandlingResult,
    OverpressureError,
    aspirate_in_place,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
    DefinedErrorData,
)
from ..errors.exceptions import PipetteNotReadyToAspirateError
from ..state.update_types import CLEAR
from ..types import CurrentWell

if TYPE_CHECKING:
    from ..execution import PipettingHandler, GantryMover
    from ..resources import ModelUtils
    from ..state.state import StateView
    from ..notes import CommandNoteAdder

AspirateInPlaceCommandType = Literal["aspirateInPlace"]


class AspirateInPlaceParams(PipetteIdMixin, AspirateVolumeMixin, FlowRateMixin):
    """Payload required to aspirate in place."""

    pass


class AspirateInPlaceResult(BaseLiquidHandlingResult):
    """Result data from the execution of a AspirateInPlace command."""

    pass


_ExecuteReturn = Union[
    SuccessData[AspirateInPlaceResult],
    DefinedErrorData[OverpressureError],
]


class AspirateInPlaceImplementation(
    AbstractCommandImpl[AspirateInPlaceParams, _ExecuteReturn]
):
    """AspirateInPlace command implementation."""

    def __init__(
        self,
        pipetting: PipettingHandler,
        hardware_api: HardwareControlAPI,
        state_view: StateView,
        command_note_adder: CommandNoteAdder,
        model_utils: ModelUtils,
        gantry_mover: GantryMover,
        **kwargs: object,
    ) -> None:
        self._pipetting = pipetting
        self._state_view = state_view
        self._hardware_api = hardware_api
        self._command_note_adder = command_note_adder
        self._model_utils = model_utils
        self._gantry_mover = gantry_mover

    async def execute(self, params: AspirateInPlaceParams) -> _ExecuteReturn:
        """Aspirate without moving the pipette.

        Raises:
            TipNotAttachedError: if no tip is attached to the pipette.
            PipetteNotReadyToAspirateError: pipette plunger is not ready.
        """
        ready_to_aspirate = self._pipetting.get_is_ready_to_aspirate(
            pipette_id=params.pipetteId,
        )
        if not ready_to_aspirate:
            raise PipetteNotReadyToAspirateError(
                "Pipette cannot aspirate in place because of a previous blow out."
                " The first aspirate following a blow-out must be from a specific well"
                " so the plunger can be reset in a known safe position."
            )

        current_position = await self._gantry_mover.get_position(params.pipetteId)
        current_location = self._state_view.pipettes.get_current_location()

        result = await aspirate_in_place(
            pipette_id=params.pipetteId,
            volume=params.volume,
            flow_rate=params.flowRate,
            location_if_error={
                "retryLocation": (
                    current_position.x,
                    current_position.y,
                    current_position.z,
                )
            },
            command_note_adder=self._command_note_adder,
            pipetting=self._pipetting,
            model_utils=self._model_utils,
        )
        if isinstance(result, DefinedErrorData):
            if (
                isinstance(current_location, CurrentWell)
                and current_location.pipette_id == params.pipetteId
            ):
                return DefinedErrorData(
                    public=result.public,
                    state_update=result.state_update.set_liquid_operated(
                        labware_id=current_location.labware_id,
                        well_names=self._state_view.geometry.get_wells_covered_by_pipette_with_active_well(
                            current_location.labware_id,
                            current_location.well_name,
                            params.pipetteId,
                        ),
                        volume_added=CLEAR,
                    ),
                    state_update_if_false_positive=result.state_update_if_false_positive,
                )
            else:
                return result
        else:
            if (
                isinstance(current_location, CurrentWell)
                and current_location.pipette_id == params.pipetteId
            ):
                return SuccessData(
                    public=AspirateInPlaceResult(volume=result.public.volume),
                    state_update=result.state_update.set_liquid_operated(
                        labware_id=current_location.labware_id,
                        well_names=self._state_view.geometry.get_wells_covered_by_pipette_with_active_well(
                            current_location.labware_id,
                            current_location.well_name,
                            params.pipetteId,
                        ),
                        volume_added=-result.public.volume
                        * self._state_view.geometry.get_nozzles_per_well(
                            current_location.labware_id,
                            current_location.well_name,
                            params.pipetteId,
                        ),
                    ),
                )
            else:
                return SuccessData(
                    public=AspirateInPlaceResult(volume=result.public.volume),
                    state_update=result.state_update,
                )


class AspirateInPlace(
    BaseCommand[AspirateInPlaceParams, AspirateInPlaceResult, OverpressureError]
):
    """AspirateInPlace command model."""

    commandType: AspirateInPlaceCommandType = "aspirateInPlace"
    params: AspirateInPlaceParams
    result: Optional[AspirateInPlaceResult] = None

    _ImplementationCls: Type[
        AspirateInPlaceImplementation
    ] = AspirateInPlaceImplementation


class AspirateInPlaceCreate(BaseCommandCreate[AspirateInPlaceParams]):
    """AspirateInPlace command request model."""

    commandType: AspirateInPlaceCommandType = "aspirateInPlace"
    params: AspirateInPlaceParams

    _CommandCls: Type[AspirateInPlace] = AspirateInPlace



"""Robot movement commands."""

from .move_to import (
    MoveTo,
    MoveToCreate,
    MoveToParams,
    MoveToResult,
    MoveToCommandType,
)
from .move_axes_to import (
    MoveAxesTo,
    MoveAxesToCreate,
    MoveAxesToParams,
    MoveAxesToResult,
    MoveAxesToCommandType,
)
from .move_axes_relative import (
    MoveAxesRelative,
    MoveAxesRelativeCreate,
    MoveAxesRelativeParams,
    MoveAxesRelativeResult,
    MoveAxesRelativeCommandType,
)
from .open_gripper_jaw import (
    openGripperJaw,
    openGripperJawCreate,
    openGripperJawParams,
    openGripperJawResult,
    openGripperJawCommandType,
)
from .close_gripper_jaw import (
    closeGripperJaw,
    closeGripperJawCreate,
    closeGripperJawParams,
    closeGripperJawResult,
    closeGripperJawCommandType,
)

__all__ = [
    # robot/moveTo
    "MoveTo",
    "MoveToCreate",
    "MoveToParams",
    "MoveToResult",
    "MoveToCommandType",
    # robot/moveAxesTo
    "MoveAxesTo",
    "MoveAxesToCreate",
    "MoveAxesToParams",
    "MoveAxesToResult",
    "MoveAxesToCommandType",
    # robot/moveAxesRelative
    "MoveAxesRelative",
    "MoveAxesRelativeCreate",
    "MoveAxesRelativeParams",
    "MoveAxesRelativeResult",
    "MoveAxesRelativeCommandType",
    # robot/openGripperJaw
    "openGripperJaw",
    "openGripperJawCreate",
    "openGripperJawParams",
    "openGripperJawResult",
    "openGripperJawCommandType",
    # robot/closeGripperJaw
    "closeGripperJaw",
    "closeGripperJawCreate",
    "closeGripperJawParams",
    "closeGripperJawResult",
    "closeGripperJawCommandType",
]



from typing import Protocol

from .types import (
    StackerAxis,
    PlatformStatus,
    Direction,
    MoveParams,
    StackerInfo,
    LEDColor,
)


class AbstractStackerDriver(Protocol):
    """Protocol for the Stacker driver."""

    async def connect(self) -> None:
        """Connect to stacker."""
        ...

    async def disconnect(self) -> None:
        """Disconnect from stacker."""
        ...

    async def is_connected(self) -> bool:
        """Check connection to stacker."""
        ...

    async def update_firmware(self, firmware_file_path: str) -> None:
        """Updates the firmware on the device."""
        ...

    async def get_device_info(self) -> StackerInfo:
        """Get Device Info."""
        ...

    async def set_serial_number(self, sn: str) -> bool:
        """Set Serial Number."""
        ...

    async def stop_motors(self) -> bool:
        """Stop all motor movement."""
        ...

    async def get_limit_switch(self, axis: StackerAxis, direction: Direction) -> bool:
        """Get limit switch status.

        :return: True if limit switch is triggered, False otherwise
        """
        ...

    async def get_platform_sensor(self, direction: Direction) -> bool:
        """Get platform sensor status.

        :return: True if platform is present, False otherwise
        """
        ...

    async def get_platform_status(self) -> PlatformStatus:
        """Get platform status."""
        ...

    async def get_hopper_door_closed(self) -> bool:
        """Get whether or not door is closed.

        :return: True if door is closed, False otherwise
        """
        ...

    async def move_in_mm(
        self, axis: StackerAxis, distance: float, params: MoveParams | None = None
    ) -> bool:
        """Move axis."""
        ...

    async def move_to_limit_switch(
        self, axis: StackerAxis, direction: Direction, params: MoveParams | None = None
    ) -> bool:
        """Move until limit switch is triggered."""
        ...

    async def home_axis(self, axis: StackerAxis, direction: Direction) -> bool:
        """Home axis."""
        ...

    async def set_led(
        self, power: float, color: LEDColor | None = None, external: bool | None = None
    ) -> bool:
        """Set LED color of status bar."""
        ...
